--------------------------------------------------------------------------------
-- copyright notice:
--------------------------------------------------------------------------------
-- copyright (c) 2007 by
--
-- CESYS GmbH / GERMANY
--
-- http://www.cesys.com
--------------------------------------------------------------------------------
-- author:
--------------------------------------------------------------------------------
-- Stefan Fischer <stefan.fischer@cesys.com>
--------------------------------------------------------------------------------
-- license:
--------------------------------------------------------------------------------
-- THIS SOURCECODE IS NOT FREE! IT IS FOR USE TOGETHER WITH THE CESYS EFM01
-- USB CARD (ARTICLE-NR.: C1050-4107) ONLY! YOU ARE NOT ALLOWED TO MODIFY AND
-- DISTRIBUTE OR USE IT WITH ANY OTHER HARDWARE, SOFTWARE OR ANY OTHER KIND OF
-- ASIC OR PROGRAMMABLE LOGIC DESIGN WITHOUT THE EXPLICIT PERMISSION OF THE
-- COPYRIGHT HOLDER!
--------------------------------------------------------------------------------
-- disclaimer of warranty:
--------------------------------------------------------------------------------
-- THIS SOURCECODE IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL, BUT
-- THERE IS NO WARRANTY OR SUPPORT FOR THIS SOURCECODE. THE COPYRIGHT HOLDER 
-- PROVIDES THIS SOURCECODE "AS IS" WITHOUT WARRANTY OF ANY KIND,
-- EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
-- WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
-- THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THIS SOURCECODE IS WITH
-- YOU. SHOULD THIS SOURCECODE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL 
-- NECESSARY SERVICING, REPAIR OR CORRECTION.
--
-- IN NO EVENT WILL THE COPYRIGHT HOLDER BE LIABLE TO YOU FOR DAMAGES, INCLUDING
-- ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
-- USE OR INABILITY TO USE THIS SOURCECODE (INCLUDING BUT NOT LIMITED TO LOSS OF
-- DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD 
-- PARTIES OR A FAILURE OF THIS SOURCECODE TO OPERATE WITH ANY OTHER
-- SOFTWARE-PROGRAMS, HARDWARE-CIRCUITS OR ANY OTHER KIND OF ASIC OR
-- PROGRAMMABLE LOGIC DESIGN), EVEN IF THE COPYRIGHT HOLDER HAS BEEN ADVISED OF 
-- THE POSSIBILITY OF SUCH DAMAGES.
--------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.NUMERIC_STD.all;

use work.wishbone.all;


entity wb_ma_fx2 is
  generic
  (
    nr_of_irqiports : positive := 1;
    nr_of_irqoports : positive := 1;
    nr_of_dbgports : positive := 1
  );
  port
  (
    SYSCON_I : in rec_syscon_port;
    MASTER_O : out rec_master_port;
    SLAVE_I : in rec_slave_port;
    
    IRQ_I : in std_logic_vector((nr_of_irqiports-1) downto 0);
    IRQ_O : out std_logic_vector((nr_of_irqoports-1) downto 0);

    fx2_ifclk_i : in std_logic;
    fx2_flag_n_i : in std_logic_vector(2 downto 0);
    fx2_sloe_n_o : out std_logic;
    fx2_slrd_n_o : out std_logic;
    fx2_slwr_n_o : out std_logic;
    fx2_pktend_n_o : out std_logic;
    fx2_fifoadr_o : out std_logic_vector(1 downto 0);
    fx2_fd_io : inout std_logic_vector(15 downto 0);

    debug : inout std_logic_vector((nr_of_dbgports-1) downto 0)
  );
end wb_ma_fx2;


architecture RTL of wb_ma_fx2 is

  signal wishbone : rec_wishbone_signal := wishbone_default;

  signal irqi : std_logic_vector((nr_of_irqiports-1) downto 0) := (others => '0');
  signal irqo : std_logic_vector((nr_of_irqoports-1) downto 0) := (others => '0');

  signal fx2_ifclk : std_logic := '0';
  signal fx2_flag_n : std_logic_vector(2 downto 0) := (others => '1');
  signal fx2_sloe_n : std_logic := '1';
  signal fx2_slrd_n : std_logic := '1';
  signal fx2_slwr_n : std_logic := '1';
  signal fx2_pktend_n : std_logic := '1';
  signal fx2_fifoadr : std_logic_vector(1 downto 0) := (others => '0');

  component fx2_slfifo_ctrl is
    generic
    (
      use_xil_dram : boolean := false;
      nr_of_dbgports : positive := 1
    );
    port
    (
      rst_i : in std_logic;
      
      -- ports connected to cypress fx2 device
      fx2_ifclk_i : in std_logic;
      fx2_flag_n_i : in std_logic_vector(2 downto 0);
      fx2_sloe_n_o : out std_logic;
      fx2_slrd_n_o : out std_logic;
      fx2_slwr_n_o : out std_logic;
      fx2_pktend_n_o : out std_logic;
      fx2_fifoadr_o : out std_logic_vector(1 downto 0);
      fx2_fd_io : inout std_logic_vector(15 downto 0);

      -- ports connected to user application
      app_fifo_wr_i : in std_logic;
      app_fifo_wr_data_i : in std_logic_vector(15 downto 0);
      app_fifo_wr_full_o : out std_logic;
      app_fifo_wr_count_o : out std_logic_vector(3 downto 0);
      app_usb_p2h_pktcount_o : out std_logic_vector(7 downto 0);

      app_fifo_rd_i : in std_logic;
      app_fifo_rd_data_o : out std_logic_vector(15 downto 0);
      app_fifo_rd_empty_o : out std_logic;
      app_fifo_rd_count_o : out std_logic_vector(3 downto 0);
      app_usb_h2p_pktcount_o : out std_logic_vector(7 downto 0);
          
      debug : inout std_logic_vector((nr_of_dbgports-1) downto 0)
    );
  end component;
    
  signal app_fifo_wr : std_logic := '0';
  signal app_fifo_wr_data : std_logic_vector(15 downto 0) := (others => '0');
  signal app_fifo_wr_full : std_logic := '0';
  signal app_fifo_wr_count : std_logic_vector(3 downto 0) := (others => '0');
  signal app_usb_p2h_pktcount : std_logic_vector(7 downto 0) := (others => '0');

  signal app_fifo_rd : std_logic := '0';
  signal app_fifo_rd_data : std_logic_vector(15 downto 0) := (others => '0');
  signal app_fifo_rd_empty : std_logic := '0';
  signal app_fifo_rd_count : std_logic_vector(3 downto 0) := (others => '0');
  signal app_usb_h2p_pktcount : std_logic_vector(7 downto 0) := (others => '0');

  component sfifo_hd_a1Kx18b0K5x36 is
    generic
    (
      nr_of_dbgports : positive := 1
    );
    port
    (
      rst : in std_logic;
      clk : in std_logic;
      
      dir_a2b_b2a_n_i : in std_logic;
  
      a_wr_i : in std_logic;
      a_wr_data_i : in std_logic_vector(17 downto 0);
      a_rd_i : in std_logic;
      a_rd_data_o : out std_logic_vector(17 downto 0);
  
      b_wr_i : in std_logic;
      b_wr_data_i : in std_logic_vector(35 downto 0);
      b_rd_i : in std_logic;
      b_rd_data_o : out std_logic_vector(35 downto 0);
  
      full_o : out std_logic;
      empty_o : out std_logic;
      count_o : out std_logic_vector(8 downto 0);
  
      debug : inout std_logic_vector((nr_of_dbgports-1) downto 0)
    );
  end component;

  signal rst : std_logic := '1';
  signal clk : std_logic := '0';

  signal fifo_dir_a2b_b2a_n : std_logic_vector(1 downto 0) := (others => '0');

  type arr_fifo_a_data is array(natural range <>) of std_logic_vector(17 downto 0);
  signal fifo_a_wr : std_logic_vector(1 downto 0) := (others => '0');
  signal fifo_a_wr_data : arr_fifo_a_data(1 downto 0) := (others => (others => '0'));
  signal fifo_a_rd : std_logic_vector(1 downto 0) := (others => '0');
  signal fifo_a_rd_data : arr_fifo_a_data(1 downto 0) := (others => (others => '0'));

  type arr_fifo_b_data is array(natural range <>) of std_logic_vector(35 downto 0);
  signal fifo_b_wr : std_logic_vector(1 downto 0) := (others => '0');
  signal fifo_b_wr_data : arr_fifo_b_data(1 downto 0) := (others => (others => '0'));
  signal fifo_b_rd : std_logic_vector(1 downto 0) := (others => '0');
  signal fifo_b_rd_data : arr_fifo_b_data(1 downto 0) := (others => (others => '0'));

  signal fifo_full : std_logic_vector(1 downto 0) := (others => '1');
  signal fifo_empty : std_logic_vector(1 downto 0) := (others => '1');
  type arr_fifo_count is array(natural range <>) of std_logic_vector(8 downto 0);
  signal fifo_count : arr_fifo_count(1 downto 0) := (others => (others => '0'));
  
  constant H2P : natural := 0;
  constant P2H : natural := 1;

  signal h2p_dval : std_logic := '0';
  signal h2p_rd_en : std_logic := '0';
  signal p2h_wr_en : std_logic := '0';
  signal addr_inc : std_logic := '0';
  
  signal transfer_count : unsigned(21 downto 0) := (others => '0');
  signal h2p_count : unsigned(6 downto 0) := (others => '0');
  signal p2h_count : unsigned(6 downto 0) := (others => '0');
  
  type t_states is
  (
    S_DECODE,
    S_ADDRESS,
    S_USB2WB,
    S_USB2WB_ACK,
    S_USB2WB_DUMMY,
    S_WB2USB,
    S_WB2USB_ACK,
    S_WB2USB_DUMMY,
    S_HEADER_DUMMY
  );
  signal state : t_states := S_DECODE;

begin

  -- connect ports to internal signals
  ------------------------------------

  -- ports of WISHBONE bus
  wishbone.syscon <= SYSCON_I;
  MASTER_O <= wishbone.master;
  wishbone.slave <= SLAVE_I;

  -- ports of interrupt bus
  irqi <= IRQ_I;
  IRQ_O <= irqo;

  -- ports of slave fifo interface
  fx2_ifclk <= fx2_ifclk_i;
  fx2_flag_n <= fx2_flag_n_i;
  fx2_sloe_n_o <= fx2_sloe_n;
  fx2_slrd_n_o <= fx2_slrd_n;
  fx2_slwr_n_o <= fx2_slwr_n;
  fx2_pktend_n_o <= fx2_pktend_n;
  fx2_fifoadr_o <= fx2_fifoadr;

  debug <=
  (
    others => 'Z'
  );

  -- instantiate low level slave fifo interface controller
  --------------------------------------------------------

  inst_fx2_slfifo_ctrl : fx2_slfifo_ctrl
  generic map
  (
    use_xil_dram => true,
    nr_of_dbgports => 1
  )
  port map
  (
    rst_i => rst,
    
    -- ports connected to cypress fx2 device
    fx2_ifclk_i => fx2_ifclk,
    fx2_flag_n_i => fx2_flag_n,
    fx2_sloe_n_o => fx2_sloe_n,
    fx2_slrd_n_o => fx2_slrd_n,
    fx2_slwr_n_o => fx2_slwr_n,
    fx2_pktend_n_o => fx2_pktend_n,
    fx2_fifoadr_o => fx2_fifoadr,
    fx2_fd_io => fx2_fd_io,

    -- ports connected to user application
    app_fifo_wr_i => app_fifo_wr,
    app_fifo_wr_data_i => app_fifo_wr_data,
    app_fifo_wr_full_o => app_fifo_wr_full,
    app_fifo_wr_count_o => app_fifo_wr_count,
    app_usb_p2h_pktcount_o => app_usb_p2h_pktcount,

    app_fifo_rd_i => app_fifo_rd,
    app_fifo_rd_data_o => app_fifo_rd_data,
    app_fifo_rd_empty_o => app_fifo_rd_empty,
    app_fifo_rd_count_o => app_fifo_rd_count,
    app_usb_h2p_pktcount_o => app_usb_h2p_pktcount, 
    
    debug => open
  );

  rst <= wishbone.syscon.rst;
  clk <= wishbone.syscon.clk;

  fifo_a_wr_data(H2P)(15 downto 0) <= app_fifo_rd_data;
  app_fifo_wr_data <= fifo_a_rd_data(P2H)(15 downto 0);
  
  -- fifo cascading logic
  -- slave fifo bus and wishbone system are fully synchronous!
  process(rst, clk, fifo_count, app_fifo_rd_empty, app_fifo_wr_count, fifo_empty)
  begin
    if (unsigned(fifo_count(H2P)) < 510) and (app_fifo_rd_empty = '0') then
      app_fifo_rd <= '1';
    else
      app_fifo_rd <= '0';
    end if;
    if (unsigned(app_fifo_wr_count) < 14) and (fifo_empty(P2H) = '0') then
      fifo_a_rd(P2H) <= '1';
    else
      fifo_a_rd(P2H) <= '0';
    end if;
    if rst = '1' then
      fifo_a_wr(H2P) <= '0';
      app_fifo_wr <= '0';
    elsif clk'event and clk = '1' then
      fifo_a_wr(H2P) <= app_fifo_rd;
      app_fifo_wr <= fifo_a_rd(P2H);
    end if;
  end process;

  conversion_16_32_bit:
  for i in 0 to 1 generate
    inst_sfifo_hd_a1Kx18b0K5x36 : sfifo_hd_a1Kx18b0K5x36
    generic map
    (
      nr_of_dbgports => 1
    )
    port map
    (
      rst => rst,
      clk => clk,

      dir_a2b_b2a_n_i => fifo_dir_a2b_b2a_n(i),

      a_wr_i => fifo_a_wr(i),
      a_wr_data_i => fifo_a_wr_data(i),
      a_rd_i => fifo_a_rd(i),
      a_rd_data_o => fifo_a_rd_data(i),

      b_wr_i => fifo_b_wr(i),
      b_wr_data_i => fifo_b_wr_data(i),
      b_rd_i => fifo_b_rd(i),
      b_rd_data_o => fifo_b_rd_data(i),

      full_o => fifo_full(i),
      empty_o => fifo_empty(i),
      count_o => fifo_count(i),

      debug => open
    );
  end generate;
    
  -- half duplex fifos have a fixed direction in this application
  fifo_dir_a2b_b2a_n(H2P) <= '1';
  fifo_dir_a2b_b2a_n(P2H) <= '0';

  -- architecture implementation
  ------------------------------

  -- as long as the usb host is the one and only master at the on-chip-bus and cyc is not
  -- needed for arbitration purposes, this signal could be statically set to '1'
  wishbone.master.cyc <= '1';

  fifo_b_wr_data(P2H)(31 downto 0) <= wishbone.slave.dat;
  wishbone.master.dat <= fifo_b_rd_data(H2P)(31 downto 0);
  
  fifo_b_rd(H2P) <= h2p_rd_en and (not fifo_empty(H2P));
  fifo_b_wr(P2H) <= (p2h_wr_en or (wishbone.slave.ack and (not wishbone.master.we))) and (not fifo_full(P2H));
  
  process(wishbone)
  begin
    if wishbone.syscon.rst = '1' then
      h2p_dval <= '0';
      h2p_rd_en <= '1';
      h2p_count <= (others => '0');
      p2h_wr_en <= '0';
      p2h_count <= (others => '0');
      state <= S_DECODE;
      wishbone.master.stb <= '0';
    elsif wishbone.syscon.clk'event and wishbone.syscon.clk = '1' then
      -- obligatory one clk cycle delay for fpga fifos
      h2p_dval <= fifo_b_rd(H2P);
      -- needed for 512 byte bulk packet alignment, meaning 128 32-bit word
      -- alignment => these are word counters not byte counters!
      if fifo_b_rd(H2P) = '1' then
        h2p_count <= h2p_count + 1; 
      end if;
      if fifo_b_wr(P2H) = '1' then
        p2h_count <= p2h_count + 1; 
      end if;
      -- fsm decoding transfer protocol and handling wishbone interface
      case state is
        -- getting direction, size and transfer mode
        when S_DECODE =>
          if h2p_dval = '1' then
            wishbone.master.we <= fifo_b_rd_data(H2P)(31);
            addr_inc <= fifo_b_rd_data(H2P)(30);
            -- transfer_count counts wishbone transactions not byte transfers
            -- => shift 2 bits
            transfer_count <= unsigned(fifo_b_rd_data(H2P)(23 downto 2));
            state <= S_ADDRESS;
          end if;
        -- getting start address
        when S_ADDRESS => 
          if h2p_dval = '1' then
            -- note that the software uses byte addressing, while the wishbone system uses
            -- 32-bit-word-addressing! so the address is shifted by two.
            wishbone.master.adr <= b"00" & fifo_b_rd_data(H2P)(31 downto 2);
            state <= S_HEADER_DUMMY;
          end if;
        -- reading don't cares from pipeline
        -- data direction decision point
        when S_HEADER_DUMMY =>
          if (fifo_b_rd(H2P) = '1') and (unsigned(h2p_count) = 127) then
            if wishbone.master.we = '1' then
              state <= S_USB2WB;
            else
              h2p_rd_en <= '0';
              state <= S_WB2USB;
            end if;
          end if;
        -- transfering data usb => wishbone
        when S_USB2WB =>
          if fifo_b_rd(H2P) = '1' then
            h2p_rd_en <= '0';
            wishbone.master.stb <= '1';
            state <= S_USB2WB_ACK;
          end if;
        when S_USB2WB_ACK =>
          if wishbone.slave.ack = '1' then
            h2p_rd_en <= '1';
            wishbone.master.stb <= '0';
            transfer_count <= transfer_count - 1;
            if addr_inc = '1' then
              wishbone.master.adr <= b"00" & std_logic_vector(unsigned(wishbone.master.adr(29 downto 0)) + 1);
            end if;
            if transfer_count = 1 then
              if unsigned(h2p_count) = 0 then
                state <= S_DECODE;
              else
                state <= S_USB2WB_DUMMY;
              end if;
            else
              state <= S_USB2WB;
            end if;
          end if;
        -- reading dummy data from pipeline
        when S_USB2WB_DUMMY =>
          if (h2p_dval = '1') and (unsigned(h2p_count) = 0) then
            state <= S_DECODE;
          end if;
        -- transfering data wishbone => usb 
        when S_WB2USB =>
          if fifo_full(P2H) = '0' then
            wishbone.master.stb <= '1';
            state <= S_WB2USB_ACK;
          end if;
        when S_WB2USB_ACK =>
          if wishbone.slave.ack = '1' then
            wishbone.master.stb <= '0';
            transfer_count <= transfer_count - 1;
            if addr_inc = '1' then
              wishbone.master.adr <= b"00" & std_logic_vector(unsigned(wishbone.master.adr(29 downto 0)) + 1);
            end if;
            if transfer_count = 1 then
              if unsigned(p2h_count) = 127 then
                h2p_rd_en <= '1';
                state <= S_DECODE;
              else
                p2h_wr_en <= '1';
                state <= S_WB2USB_DUMMY;
              end if;
            else
              state <= S_WB2USB;
            end if;
          end if;
        -- adding dummy data to pipeline
        when S_WB2USB_DUMMY =>
          if (fifo_b_wr(P2H) = '1') and (unsigned(p2h_count) = 127) then
            h2p_rd_en <= '1';
            p2h_wr_en <= '0';
            state <= S_DECODE;
          end if;
        when others => null;
      end case;
    end if;
  end process;
  
end RTL;